/**
 * P4
 * http://snarfed.org/space/p4
 *
 * Copyright 2001-2004
 * Ken Ashcraft <kashcraft+p4@gmail.com>
 * Ryan Barrett <p4@ryanb.org>
 * Maulik Shah <p4@maulik.net>
 * Nathan Stoll <nathanstoll+p4@gmail.com>
 *
 * @file receiveq.h
 *
 * This data structure holds all incoming messages for plugins. They are
 * indexed plugin_id in a FIFO manner. There can only be one instance of the
 * creceiveq type since it is a singleton type.
 *
 * What do we do about short-circuiting incoming messages for a plugin that we
 * aren't running?
 *
 * some sort of method to see if plugins are not picking up their messages?
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifndef _P4_RECEIVEQ_H
#define _P4_RECEIVEQ_H

#include "protocol.h"

#include <pthread.h>
#include <sys/types.h>
#include <queue>
#include <utility>
#include <map>
using namespace std;

/************************************************************************
 * FORWARD DECLARATIONS
 ***********************************************************************/
class cmsg_data;

typedef pair<cmsg_data *, queue<cmsg_data *> > received_data;

/************************************************************************
 * CLASS creceiveq
 ***********************************************************************/
class creceiveq {

 public:
  
  /** instance
   *
   * Returns the only instance of creceiveq.  This implements the 
   * singleton pattern.
   */
  static creceiveq &instance();

  /** wipe
   *
   * Removes all entries from the receive queue.  All cmsg objects
   * will be freed.
   */
  uint wipe();

  /** plugin_loaded
   *
   * Tells the receive queue to prepare for incoming messages for
   * plugin.
   */
  uint plugin_loaded(ushort plugin_id);

  /** plugin_stopped
   *
   * Tells the receive queue that no more messages for plugin are
   * incoming.  Any enqueued messages for plugin are freed.
   */
  uint plugin_stopped(ushort plugin_id);

  /** enqueue
   *
   * Users are only interested in incoming messages of the cmsg_data
   * type.  After this call, the data from the message is accessible
   * via the get_next_message call.  If we don't know about the plugin
   * for this message, returns kerr_unknown_plugin.
   */
  uint enqueue(cmsg_data *msg);

  /** peek
   *
   * Gives the size of the next incoming message for plugin.
   */
  uint peek(ushort plugin_id, int &size) const;

  /** dequeue
   *
   * Gets the next message for plugin.  This function is closely tied
   * with the receive function in p4plugin.h.
   *
   * Writes the name of the user who sent the message to the array
   * that user points to message. The parameter size specifies the
   * size of the data buffer.  If size is smaller than the size of the 
   * next message, returns an error.  The actual amount copied into
   * data is written to size_written. encrypted tells whether the message
   * was encrypted when it arrived.
   */
  uint dequeue(ushort plugin_id, char *user, char *data, int size,
               int &size_written, bool &encrypted);

  /**
   * Authenticates that the last message received by the calling plugin was
   * actually sent by the said user. If the signature is valid, true is written
   * to authenticated, false otherwise.
   */
  uint authenticate(ushort plugin_id, bool &authenticated) const;

  /** num_plugins
   * 
   * Returns the number of plugins that are loaded.
   */
  ushort num_plugins() const;

  /** get_plugins
   * 
   * Writes the active plugins to plugins and writes the 
   * num_written.  If num_plugins is smaller than the number
   * of active plugins, returns kerr_buffer_too_small.
   */
  uint get_plugins(ushort num_plugins, plugin_info *plugins, 
                   ushort &num_written) const;

  protected:

  creceiveq();
  
  ~creceiveq();

  /************************************************************************
   * CONSTANTS AND MEMBERS
   ***********************************************************************/
  
  /** queues
   *
   * For each active plugin, there is an entry in queues.  Each 
   * entry is indexed by its plugin_id and contains a pair.  The
   * pair holds a reference to the last message read by the plugin
   * and a deque of all messages waiting to be read by the plugin.
   */
  map<ushort, received_data> queues;

  mutable pthread_mutex_t mtx;
};


/***************************************************************************
 * INLINE METHODS
 **************************************************************************/

inline creceiveq &creceiveq::instance()
{
  static creceiveq rq;

  return rq;
}

#endif  // _P4_RECEIVEQ_H

